#!/usr/bin/env python3

"""frccontrol example for a differential drive."""

import sys

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

import frccontrol as fct

if "--noninteractive" in sys.argv:
    mpl.use("svg")


class DifferentialDrive:
    """An frccontrol system representing a differential drive."""

    def __init__(self, dt):
        """DifferentialDrive subsystem.

        Keyword arguments:
        dt -- time between model/controller updates
        """
        self.dt = dt

        self.in_low_gear = False

        # Number of motors per side
        num_motors = 2.0

        # High and low gear ratios of differential drive
        Glow = 60.0 / 11.0
        Ghigh = 60.0 / 11.0

        # Drivetrain mass in kg
        m = 52
        # Radius of wheels in meters
        r = 0.08255 / 2.0
        # Radius of robot in meters
        rb = 0.59055 / 2.0
        # Moment of inertia of the differential drive in kg-m^2
        J = 6.0

        # Gear ratios of left and right sides of differential drive respectively
        if self.in_low_gear:
            Gl = Glow
            Gr = Glow
        else:
            Gl = Ghigh
            Gr = Ghigh

        self.plant = fct.models.differential_drive(
            fct.models.MOTOR_CIM, num_motors, m, r, rb, J, Gl, Gr
        )

        # Sim variables
        self.sim = self.plant.to_discrete(self.dt)
        self.x = np.zeros((2, 1))
        self.u = np.zeros((2, 1))
        self.y = np.zeros((2, 1))

        # States: left velocity (m/s), right velocity (m/s)
        # Inputs: left voltage (V), right voltage (V)
        # Outputs: left velocity (m/s), right velocity (m/s)
        self.observer = fct.KalmanFilter(self.plant, [1.0, 1.0], [0.01, 0.01], self.dt)
        self.feedforward = fct.LinearPlantInversionFeedforward(
            self.plant.A, self.plant.B, self.dt
        )
        if self.in_low_gear:
            self.feedback = fct.LinearQuadraticRegulator(
                self.plant.A,
                self.plant.B,
                [1.0, 1.0],
                [12.0, 12.0],
                self.dt,
            )
        else:
            self.feedback = fct.LinearQuadraticRegulator(
                self.plant.A,
                self.plant.B,
                [0.95, 0.95],
                [12.0, 12.0],
                self.dt,
            )

        self.u_min = np.array([[-12.0], [-12.0]])
        self.u_max = np.array([[12.0], [12.0]])

    def update(self, r, next_r):
        """
        Advance the model by one timestep.

        Keyword arguments:
        r -- the current reference
        next_r -- the next reference
        """
        # Update sim model
        self.x = self.sim.A @ self.x + self.sim.B @ self.u
        self.y = self.sim.C @ self.x + self.sim.D @ self.u

        self.observer.predict(self.u, self.dt)
        self.observer.correct(self.u, self.y)
        self.u = np.clip(
            self.feedforward.calculate(next_r)
            + self.feedback.calculate(self.observer.x_hat, r),
            self.u_min,
            self.u_max,
        )


def main():
    """Entry point."""

    dt = 0.005
    diff_drive = DifferentialDrive(dt)

    ts, _, vprof, _ = fct.generate_trapezoid_profile(
        max_v=3.5, time_to_max_v=1.0, dt=dt, goal=50.0
    )

    # Run simulation
    refs = [np.array([[vprof[i]], [vprof[i]]]) for i in range(len(ts))]
    r_rec, x_rec, u_rec, _ = fct.generate_time_responses(diff_drive, refs)

    fct.plot_time_responses(
        [
            "Left velocity (m/s)",
            "Right velocity (m/s)",
        ],
        ["Left voltage (V)", "Right voltage (V)"],
        ts,
        r_rec,
        x_rec,
        u_rec,
    )
    if "--noninteractive" in sys.argv:
        plt.savefig("differential_drive_response.svg")
    else:
        plt.show()


if __name__ == "__main__":
    main()



"""
Extended Kalman filter class.
"""

import numpy as np
import scipy as sp

from .ctrlutil import make_cov_matrix, obsv
from .discretization import discretize_aq, discretize_r
from .numerical_integration import rkdp
from .numerical_jacobian import numerical_jacobian_x


class ExtendedKalmanFilter:
    """
    Extended Kalman filter class.
    """

    def __init__(
        self,
        states,
        inputs,
        f,
        h,
        state_std_devs: list[float],
        measurement_std_devs: list[float],
        dt: float,
    ):
        """
        Constructs an extended Kalman filter.

        Keyword arguments:
        states -- Number of states.
        inputs -- Number of inputs.
        f -- A vector-valued function of x and u that returns the derivative of
             the state vector.
        h -- A vector-valued function of x and u that returns the measurement
             vector.
        state_std_devs -- Standard deviations of model states.
        measurement_std_devs -- Standard deviations of measurements.
        dt -- Nominal discretization timestep.
        """
        self.states = states
        self.inputs = inputs
        self.outputs = len(measurement_std_devs)

        self.f = f
        self.h = h
        self.dt = dt
        self.x_hat = np.zeros((self.states, 1))

        self.contQ = make_cov_matrix(state_std_devs)
        self.contR = make_cov_matrix(measurement_std_devs)

        contA = numerical_jacobian_x(
            self.states, self.states, self.f, self.x_hat, np.zeros((inputs, 1))
        )
        C = numerical_jacobian_x(
            self.outputs, self.states, self.h, self.x_hat, np.zeros((inputs, 1))
        )

        discA, discQ = discretize_aq(contA, self.contQ, dt)
        discR = discretize_r(self.contR, dt)

        if (
            np.linalg.matrix_rank(obsv(discA, C)) == self.states
            and self.outputs <= self.states
        ):
            self.init_P = sp.linalg.solve_discrete_are(
                a=discA.T, b=C.T, q=discQ, r=discR
            )
        else:
            self.init_P = np.zeros((states, states))
        self.P = self.init_P

    def reset(self):
        """
        Resets the observer.
        """
        self.x_hat = np.zeros(self.x_hat.shape)
        self.P = self.init_P

    def predict(self, u, dt):
        """
        Project the model into the future with a new control input u.

        Keyword arguments:
        u -- New control input from controller.
        dt -- Timestep for prediction.
        """
        # Find continuous A
        contA = numerical_jacobian_x(
            self.states,
            self.states,
            self.f,
            self.x_hat,
            u,
        )

        # Find discrete A and Q
        discA, discQ = discretize_aq(contA, self.contQ, dt)

        self.x_hat = rkdp(self.f, self.x_hat, u, dt)

        # Pₖ₊₁⁻ = APₖ⁻Aᵀ + Q
        self.P = discA @ self.P @ discA.T + discQ

        self.dt = dt

    def correct(self, u, y):
        """
        Correct the state estimate x-hat using the measurements in y.

        Keyword arguments:
        u -- Same control input used in the last predict step.
        y -- Measurement vector.
        """
        C = numerical_jacobian_x(self.outputs, self.states, self.h, self.x_hat, u)
        R = discretize_r(self.contR, self.dt)

        S = C @ self.P @ C.T + R

        # We want to put K = PCᵀS⁻¹ into Ax = b form so we can solve it more
        # efficiently.
        #
        # K = PCᵀS⁻¹
        # KS = PCᵀ
        # (KS)ᵀ = (PCᵀ)ᵀ
        # SᵀKᵀ = CPᵀ
        #
        # The solution of Ax = b can be found via x = A.solve(b).
        #
        # Kᵀ = Sᵀ.solve(CPᵀ)
        # K = (Sᵀ.solve(CPᵀ))ᵀ
        K = np.linalg.solve(S.T, C @ self.P.T).T

        # x̂ₖ₊₁⁺ = x̂ₖ₊₁⁻ + Kₖ₊₁(y − h(x̂ₖ₊₁⁻, uₖ₊₁))
        self.x_hat += K @ (y - self.h(self.x_hat, u))

        # Pₖ₊₁⁺ = (I−Kₖ₊₁C)Pₖ₊₁⁻(I−Kₖ₊₁C)ᵀ + Kₖ₊₁RKₖ₊₁ᵀ
        # Use Joseph form for numerical stability
        self.P = (np.eye(self.states) - K @ C) @ self.P @ (
            np.eye(self.states) - K @ C
        ).T + K @ R @ K.T



#!/usr/bin/env python3

"""frccontrol example for a double-jointed arm."""

import sys

import matplotlib as mpl
from matplotlib import animation
import matplotlib.pyplot as plt
import numpy as np

import frccontrol as fct

if "--noninteractive" in sys.argv:
    mpl.use("svg")


class DoubleJointedArm:
    """
    An frccontrol system representing a double-jointed arm.

    States: [joint angle 1, joint angle 2,
             joint angular velocity 1, joint angular velocity 2,
             input error 1, input error 2]
    Inputs: [voltage 1, voltage 2]
    """

    def __init__(self, dt):
        """
        Double-jointed arm subsystem.

        Keyword arguments:
        dt -- simulation step time
        """
        self.dt = dt

        self.constants = DoubleJointedArmConstants()

        q_pos = 0.01745
        q_vel = 0.1745329
        q_error = 10
        r_pos = 0.05
        self.observer = fct.ExtendedKalmanFilter(
            6,
            2,
            self.f,
            self.h,
            [q_pos, q_pos, q_vel, q_vel, q_error, q_error],
            [r_pos, r_pos],
            self.dt,
        )

        self.x = np.zeros((6, 1))
        self.u = np.zeros((2, 1))
        self.y = np.zeros((2, 1))

        self.u_min = np.array([[-12.0]])
        self.u_max = np.array([[12.0]])

    # pragma pylint: disable=unused-argument
    def update(self, r, next_r):
        """
        Advance the model by one timestep.

        Keyword arguments:
        r -- the current reference
        next_r -- the next reference
        """
        self.x = fct.rkdp(self.f, self.x, self.u, self.dt)
        self.y = self.h(self.x, self.u)
        # self.y += np.array(
        #     [np.random.multivariate_normal(mean=[0, 0], cov=np.diag([1e-4, 1e-4]))]
        # ).T

        self.observer.predict(self.u, self.dt)
        self.observer.correct(self.u, self.y)

        rdot = (next_r - r) / self.dt

        u_ff = self.feedforward(r, rdot)

        A = fct.numerical_jacobian_x(
            6,
            6,
            self.f,
            self.observer.x_hat,
            self.feedforward(self.observer.x_hat, rdot),
        )
        B = fct.numerical_jacobian_u(
            6,
            2,
            self.f,
            self.observer.x_hat,
            self.feedforward(self.observer.x_hat, rdot),
        )
        u_fb = fct.LinearQuadraticRegulator(
            A[:4, :4],
            B[:4, :],
            [0.01745, 0.01745, 0.08726, 0.08726],
            [12.0, 12.0],
            self.dt,
        ).K @ (r[0:4] - self.observer.x_hat[0:4])

        # Voltage output from input error estimation. The estimate is in
        # Newton-meters, so has to be converted to volts.
        u_err = np.linalg.solve(self.constants.B, self.observer.x_hat[4:])

        self.u = np.clip(u_ff + u_fb - u_err, self.u_min, self.u_max)

    def get_dynamics_matrices(self, x):
        """Gets the dynamics matrices for the given state.

        See derivation at:
        https://www.chiefdelphi.com/t/whitepaper-two-jointed-arm-dynamics/423060

        Keyword arguments:
        x -- current system state
        """
        theta1, theta2, omega1, omega2 = x[:4].flat
        c2 = np.cos(theta2)

        l1 = self.constants.l1
        r1 = self.constants.r1
        r2 = self.constants.r2
        m1 = self.constants.m1
        m2 = self.constants.m2
        I1 = self.constants.I1
        I2 = self.constants.I2
        g = self.constants.g

        hM = l1 * r2 * c2
        M = (
            m1 * np.array([[r1 * r1, 0], [0, 0]])
            + m2
            * np.array(
                [[l1 * l1 + r2 * r2 + 2 * hM, r2 * r2 + hM], [r2 * r2 + hM, r2 * r2]]
            )
            + I1 * np.array([[1, 0], [0, 0]])
            + I2 * np.array([[1, 1], [1, 1]])
        )

        hC = -m2 * l1 * r2 * np.sin(theta2)
        C = np.array([[hC * omega2, hC * omega1 + hC * omega2], [-hC * omega1, 0]])

        G = (
            g * np.cos(theta1) * np.array([[m1 * r1 + m2 * self.constants.l1, 0]]).T
            + g * np.cos(theta1 + theta2) * np.array([[m2 * r2, m2 * r2]]).T
        )

        return M, C, G

    def f(self, x, u):
        """
        Dynamics model.

        Keyword arguments:
        x -- state vector
        u -- input vector
        """
        M, C, G = self.get_dynamics_matrices(x)

        omega = x[2:4]

        # Motor dynamics
        torque = self.constants.A @ x[2:4] + self.constants.B @ u

        # dx/dt = [ω α 0]ᵀ
        return np.block(
            [[x[2:4]], [np.linalg.solve(M, torque - C @ omega - G)], [np.zeros((2, 1))]]
        )

    def h(self, x, u):
        """
        Measurement model.

        Keyword arguments:
        x -- state vector
        u -- input vector
        """
        return x[:2, :]

    def feedforward(self, x, xdot):
        """
        Arm feedforward.
        """
        M, C, G = self.get_dynamics_matrices(x)
        omega = x[2:4]
        return np.linalg.solve(
            self.constants.B, M @ xdot[2:4] + C @ omega + G - self.constants.A @ omega
        )


class DoubleJointedArmConstants:
    """
    Double-jointed arm model constants.
    """

    def __init__(self):
        # Length of segments
        self.l1 = 46.25 * 0.0254
        self.l2 = 41.80 * 0.0254

        # Mass of segments
        self.m1 = 9.34 * 0.4536
        self.m2 = 9.77 * 0.4536

        # Distance from pivot to CG for each segment
        self.r1 = 21.64 * 0.0254
        self.r2 = 26.70 * 0.0254

        # Moment of inertia about CG for each segment
        self.I1 = 2957.05 * 0.0254 * 0.0254 * 0.4536
        self.I2 = 2824.70 * 0.0254 * 0.0254 * 0.4536

        # Gearing of each segment
        self.G1 = 70.0
        self.G2 = 45.0

        # Number of motors in each gearbox
        self.N1 = 1
        self.N2 = 2

        # Gravity
        self.g = 9.806

        motor = fct.models.MOTOR_NEO

        # torque = A * velocity + B * voltage
        self.B = (
            np.array([[self.N1 * self.G1, 0], [0, self.N2 * self.G2]])
            * motor.Kt
            / motor.R
        )
        self.A = (
            -np.array(
                [[self.G1 * self.G1 * self.N1, 0], [0, self.G2 * self.G2 * self.N2]]
            )
            * motor.Kt
            / motor.Kv
            / motor.R
        )

    def fwd_kinematics(self, x):
        """
        Forward kinematics for the given state.
        """
        theta1, theta2 = x[:2]

        joint2 = np.array([self.l1 * np.cos(theta1), self.l1 * np.sin(theta1)])
        end_eff = joint2 + np.array(
            [self.l2 * np.cos(theta1 + theta2), self.l2 * np.sin(theta1 + theta2)]
        )

        return joint2, end_eff

    def inv_kinematics(self, x, y, invert=False):
        """
        Inverse kinematics for a target position pos (x, y). The invert flag
        controls elbow direction.

        Keyword arguments:
        x -- x position
        y -- y position

        Returns:
        theta1 -- joint 1 angle
        theta2 -- joint 2 angle
        """
        theta2 = np.arccos(
            (x * x + y * y - (self.l1 * self.l1 + self.l2 * self.l2))
            / (2 * self.l1 * self.l2)
        )

        if invert:
            theta2 = -theta2

        theta1 = np.arctan2(y, x) - np.arctan2(
            self.l2 * np.sin(theta2), self.l1 + self.l2 * np.cos(theta2)
        )

        return theta1, theta2


def main():
    """Entry point."""

    dt = 0.02

    constants = DoubleJointedArmConstants()

    def to_state(x, y, invert):
        theta1, theta2 = constants.inv_kinematics(x, y, invert)
        return np.array([[theta1], [theta2], [0], [0]])

    state1 = to_state(1.5, -1, False)
    state2 = to_state(1.5, 1, True)
    state3 = to_state(-1.8, 1, False)

    traj1 = fct.trajectory.interpolate_states(0, 3, state1, state2)
    traj2 = fct.trajectory.interpolate_states(4, 8, state2, state3)
    traj = traj1.append(traj2)

    t_rec = np.arange(0, traj.end_time + dt + 1, dt)

    # Generate references for simulation
    refs = []
    for t in t_rec:
        r = np.block([[traj.sample(t)[:4]], [np.zeros((2, 1))]])
        refs.append(r)

    double_jointed_arm = DoubleJointedArm(dt)
    double_jointed_arm.x = refs[0]
    double_jointed_arm.observer.x_hat = refs[0]

    r_rec, x_rec, u_rec, _ = fct.generate_time_responses(double_jointed_arm, refs)
    fct.plot_time_responses(
        [
            "Angle 1 (rad)",
            "Angle 2 (rad)",
            "Angular velocity 1 (rad/s)",
            "Angular velocity 2 (rad/s)",
            "Input Error 1 (N-m)",
            "Input Error 2 (N-m)",
        ],
        ["Voltage 1 (V)", "Voltage 2 (V)"],
        t_rec,
        r_rec,
        x_rec,
        u_rec,
    )
    if "--noninteractive" in sys.argv:
        plt.savefig("double_jointed_arm_response.svg")
    else:
        animate_arm(double_jointed_arm, x_rec, r_rec)


def animate_arm(arm: DoubleJointedArm, x_rec, r_rec):
    """
    Animates arm.

    Keyword arguments:
    arm -- the arm being simulated
    x_rec -- state recording
    r_rec -- reference recording
    """

    def get_arm_joints(x):
        """
        Get the x-y positions of all three robot joints:

        - base joint
        - elbow
        - end effector
        """
        joint_pos, eff_pos = arm.constants.fwd_kinematics(x)
        return np.array([0, joint_pos[0, 0], eff_pos[0, 0]]), np.array(
            [0, joint_pos[1, 0], eff_pos[1, 0]]
        )

    fig = plt.figure()
    ax = fig.add_subplot()
    ax.axis("square")

    total_len = arm.constants.l1 + arm.constants.l2
    ax.set_xlim(-total_len, total_len)
    ax.set_ylim(-total_len, total_len)

    xs, ys = get_arm_joints(x_rec[:, 0:1])
    arm_line = ax.plot(xs, ys, "-o", label="State")[0]

    xs, ys = get_arm_joints(r_rec[:, 0:1])
    ref_line = ax.plot(xs, ys, "--o", label="Reference")[0]

    ax.legend(loc="lower left")

    def animate(i):
        xs, ys = get_arm_joints(x_rec[:, i : i + 1])
        arm_line.set_data(xs, ys)

        xs, ys = get_arm_joints(r_rec[:, i : i + 1])
        ref_line.set_data(xs, ys)

        return arm_line, ref_line

    # pragma pylint: disable=unused-variable
    anim = animation.FuncAnimation(
        fig,
        animate,
        frames=x_rec.shape[1],
        interval=int(arm.dt * 1000),
        blit=True,
        repeat=True,
    )
    plt.show()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""frccontrol example for a flywheel."""

import math
import sys

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

import frccontrol as fct

if "--noninteractive" in sys.argv:
    mpl.use("svg")


class Flywheel:
    """An frccontrol system representing a flyhweel."""

    def __init__(self, dt):
        """Flywheel subsystem.

        Keyword arguments:
        dt -- time between model/controller updates
        """
        self.dt = dt

        # Number of motors
        num_motors = 1.0
        # Flywheel moment of inertia in kg-m^2
        J = 0.00032
        # Gear ratio
        G = 12.0 / 18.0
        self.plant = fct.models.flywheel(fct.models.MOTOR_775PRO, num_motors, J, G)

        # Sim variables
        self.sim = self.plant.to_discrete(self.dt)
        self.x = np.zeros((1, 1))
        self.u = np.zeros((1, 1))
        self.y = np.zeros((1, 1))

        # States: angular velocity (rad/s)
        # Inputs: voltage (V)
        # Outputs: angular velocity (rad/s)
        self.observer = fct.KalmanFilter(self.plant, [1.0], [0.01], self.dt)
        self.feedforward = fct.LinearPlantInversionFeedforward(
            self.plant.A, self.plant.B, self.dt
        )
        self.feedback = fct.LinearQuadraticRegulator(
            self.plant.A, self.plant.B, [9.42], [12.0], self.dt
        )

        self.u_min = np.array([[-12.0]])
        self.u_max = np.array([[12.0]])

    def update(self, r, next_r):
        """
        Advance the model by one timestep.

        Keyword arguments:
        r -- the current reference
        next_r -- the next reference
        """
        # Update sim model
        self.x = self.sim.A @ self.x + self.sim.B @ self.u
        self.y = self.sim.C @ self.x + self.sim.D @ self.u

        self.observer.predict(self.u, self.dt)
        self.observer.correct(self.u, self.y)
        self.u = np.clip(
            self.feedforward.calculate(next_r)
            + self.feedback.calculate(self.observer.x_hat, r),
            self.u_min,
            self.u_max,
        )


def main():
    """Entry point."""

    dt = 0.005
    flywheel = Flywheel(dt)

    # Set up graphing
    l0 = 0.1
    l1 = l0 + 5.0
    l2 = l1 + 0.1
    ts = np.arange(0, l2 + 5.0, dt)

    # Run simulation
    refs = []
    for t in ts:
        if t < l0:
            r = np.array([[0.0]])
        elif t < l1:
            r = np.array([[9000 / 60 * 2 * math.pi]])
        else:
            r = np.array([[0.0]])
        refs.append(r)
    r_rec, x_rec, u_rec, _ = fct.generate_time_responses(flywheel, refs)

    fct.plot_time_responses(
        ["Angular velocity (rad/s)"],
        ["Voltage (V)"],
        ts,
        r_rec,
        x_rec,
        u_rec,
    )
    if "--noninteractive" in sys.argv:
        plt.savefig("flywheel_response.svg")
    else:
        plt.show()


if __name__ == "__main__":
    main()



"""
Linear Kalman filter class.
"""

import numpy as np
import scipy as sp

from .ctrlutil import make_cov_matrix, obsv
from .discretization import discretize_ab, discretize_aq, discretize_r


class KalmanFilter:
    """
    Linear Kalman filter class.
    """

    def __init__(
        self,
        plant: sp.signal.StateSpace,
        state_std_devs: list[float],
        measurement_std_devs: list[float],
        dt: float,
    ):
        """
        Constructs a Kalman filter.

        Keyword arguments:
        plant -- The plant used for the prediction step.
        state_std_devs -- Standard deviations of model states.
        measurement_std_devs -- Standard deviations of measurements.
        dt -- Nominal discretization timestep.
        """
        self.plant = plant

        A, Q = discretize_aq(plant.A, make_cov_matrix(state_std_devs), dt)
        C = plant.C
        R = discretize_r(make_cov_matrix(measurement_std_devs), dt)

        if np.linalg.matrix_rank(obsv(A, C)) < A.shape[0]:
            print(f"Warning: The system is unobservable\n\nA = {A}\nC = {C}\n")

        P = sp.linalg.solve_discrete_are(a=A.T, b=C.T, q=Q, r=R)

        # S = CPCᵀ + R
        S = C @ P @ C.T + R

        # We want to put K = PCᵀS⁻¹ into Ax = b form so we can solve it more
        # efficiently.
        #
        # K = PCᵀS⁻¹
        # KS = PCᵀ
        # (KS)ᵀ = (PCᵀ)ᵀ
        # SᵀKᵀ = CPᵀ
        #
        # The solution of Ax = b can be found via x = A.solve(b).
        #
        # Kᵀ = Sᵀ.solve(CPᵀ)
        # K = (Sᵀ.solve(CPᵀ))ᵀ
        self.K = np.linalg.solve(S.T, C @ P.T).T

        self.x_hat = np.zeros((self.plant.A.shape[0], 1))

    def reset(self):
        """
        Resets the observer.
        """
        self.x_hat = np.zeros(self.x_hat.shape)

    def predict(self, u, dt):
        """
        Project the model into the future with a new control input u.

        Keyword arguments:
        u -- New control input from controller.
        dt -- Timestep for prediction.
        """
        A, B = discretize_ab(self.plant.A, self.plant.B, dt)
        self.x_hat = A @ self.x_hat + B @ u

    def correct(self, u, y):
        """
        Correct the state estimate x-hat using the measurements in y.

        Keyword arguments:
        u -- Same control input used in the last predict step.
        y -- Measurement vector.
        """
        # x̂ₖ₊₁⁺ = x̂ₖ₊₁⁻ + K(y − (Cx̂ₖ₊₁⁻ + Duₖ₊₁))
        self.x_hat += self.K @ (y - (self.plant.C @ self.x_hat + self.plant.D @ u))



"""
Cubic spline trajectory generator.
"""

import numpy as np


class Trajectory:
    """
    Cubic spline trajectory.
    """

    def __init__(self, times, states):
        """
        Initialize a Trajectory.

        Keyword arguments:
        times -- 1D array of trajectory timestamps.
        states -- 2D array of corresponding states, where each state is a column
        """
        self.times = times.flatten()
        if states.shape[1] != len(times):
            raise RuntimeError(
                f"must have same number of times and states; {len(times)} != {states.shape[1]}"
            )
        self.states = states
        self.start_time = times[0]
        self.end_time = times[-1]

    def clip_time(self, time):
        """
        Limit Trajectory timestamp between start_time and end_time.

        Keyword arguments:
        time -- the time to clip
        """
        return np.clip(time, self.start_time, self.end_time)

    def insert(self, time, state):
        """
        Insert state at the given time.

        Keyword arguments:
        time -- the time
        state -- the state
        """
        if state.ndim == 1:
            state = np.array([state]).T

        if state.shape[0] != self.states.shape[0]:
            raise RuntimeError(
                f"state must have the same number of rows; {state.shape[0]} != {self.states.shape[0]}"
            )

        if state.shape[1] != 1:
            raise RuntimeError(
                f"state must have exactly one column, had {state.shape[1]}"
            )

        before_idx_list = np.where(self.times <= time)[0]
        after_idx_list = np.where(self.times >= time)[0]

        # if there are no elements before the new time
        if before_idx_list.size == 0:
            # add to start of Trajectory
            self.times = np.insert(self.times, 0, time)
            self.states = np.insert(self.states, 0, state.T, axis=1)
            self.start_time = time
            return self
        # if there are no elements after the new time
        elif after_idx_list.size == 0:
            # add to end of Trajectory
            self.times = np.append(self.times, time)
            self.states = np.append(self.states, state, axis=1)
            self.end_time = time
            return self

        prev_idx = before_idx_list[-1]
        next_idx = after_idx_list[0]

        if self.times[prev_idx] == time:
            # time already in Trajectory; overwrite
            self.states[:, prev_idx] = state.flat
        elif self.times[next_idx] == time:
            # time already in Trajectory; overwrite
            self.states[:, next_idx] = state.flat
        else:
            # add to middle of Trajectory
            self.times = np.insert(self.times, next_idx, time)
            self.states = np.insert(self.states, next_idx, state.T, axis=1)

        return self

    def sample(self, time):
        """
        Sample the trajectory for the given time.

        Linearly interpolates between trajectory samples. If time is outside of
        trajectory, gives the start/end state.

        Keyword arguments:
        time -- time to sample
        """
        time = self.clip_time(time)
        prev_idx = np.where(self.times <= time)[0][-1]
        next_idx = np.where(self.times >= time)[0][0]

        if prev_idx == next_idx:
            return np.array([self.states[:, prev_idx]]).T

        prev_val = np.array([self.states[:, prev_idx]]).T
        next_val = np.array([self.states[:, next_idx]]).T
        prev_time = self.times[prev_idx]
        next_time = self.times[next_idx]

        return prev_val + (next_val - prev_val) / (next_time - prev_time) * (
            time - prev_time
        )

    def append(self, other):
        """
        Append another trajectory to this trajectory.

        Will adjust timestamps on the appended trajectory so it starts
        immediately after the current trajectory ends. Skips the first element
        of the other trajectory to avoid repeats.

        Keyword arguments:
        other -- The other trajectory to append to this one.
        """
        # Create new trajectory based off of this one
        combined = Trajectory(self.times, self.states)

        # Adjust timestamps on other trajectory
        other.times = other.times + combined.end_time - other.start_time

        # Combine the time and states
        combined.times = np.concatenate((combined.times, other.times[1:]))
        combined.states = np.concatenate((combined.states, other.states[:, 1:]), 1)

        # Update the end time
        combined.end_time = max(combined.times)

        return combined


def from_coeffs(coeffs: np.matrix, t_0, t_f, n=100) -> Trajectory:
    """
    Generate a trajectory from a polynomial coefficients matrix.

    Keyword arguments:
    coeffs -- Polynomial coefficients as columns in increasing order.
              Can have arbitrarily many columns.
    t_0 -- time to start the interpolation
    t_f -- time to end the interpolation
    n -- number of interpolation samples (default 100)

    Returns:
    Trajectory following the interpolation. The states will be in the form:
    [pos_1, ..., pos_n, vel_1, ..., vel_n, accel_1, ..., accel_n]
    Where n is the number of columns in coeffs.
    """
    order = np.size(coeffs, 0) - 1
    t = np.array([np.linspace(t_0, t_f, n)]).T

    pos_t_vec = np.power(t, np.arange(order + 1))
    pos_vec = pos_t_vec @ coeffs

    vel_t_vec = np.concatenate(
        (
            np.zeros((n, 1)),
            np.multiply(
                pos_t_vec[:, 0:-1], np.repeat(np.array([np.arange(order) + 1]), n, 0)
            ),
        ),
        1,
    )
    vel_vec = vel_t_vec @ coeffs

    acc_t_vec = np.concatenate(
        (
            np.zeros((n, 2)),
            np.multiply(
                vel_t_vec[:, 1:-1],
                np.repeat(np.array([np.arange(order - 1) + 2]), n, 0),
            ),
        ),
        1,
    )
    acc_vec = acc_t_vec @ coeffs

    states = np.concatenate((pos_vec, vel_vec, acc_vec), 1).T
    return Trajectory(t, states)


def interpolate_states(t_0, t_f, state_0, state_f):
    """
    Fit cubic polynomial between states.

    Keyword arguments:
    t_0 -- initial time
    t_f -- final time
    state_0 -- initial state
    state_f -- final state
    """
    return from_coeffs(__cubic_interpolation(t_0, t_f, state_0, state_f), t_0, t_f)


def __cubic_interpolation(t_0, t_f, state_0, state_f):
    """
    Perform cubic interpolation between state₀ at t = t₀ and state_f at t = t_f.
    Solves using the matrix equation:

      [1  t_0   t_0²   t_0³][c₀₁  c₀₂]   [x_0₁  x_0₂]
      [0  1    2t_0   3t_0²][c₁₁  c₁₂] = [v_0₁  v_0₂]
      [1  t_f   t_f²   t_f³][c₂₁  c₂₂]   [x_f₁  x_f₂]
      [0  1    2t_f   3t_f²][c₃₁  c₃₂]   [v_f₁  v_f₂]

    To find the cubic polynomials:

      x₁(t) = c₀₁ + c₁₁t + c₂₁t² + c₃₁t³
      x₂(t) = c₀₂ + c₁₂t + c₂₂t² + c₃₂t³

    where x₁ is the first joint position and x₂ is the second joint position,
    such that the arm is in state_0 [x_0₁, x_0₂, v_0₁, v_0₂]ᵀ at t_0 and state_f
    [x_f₁, x_f₂, v_f₁, v_f₂]ᵀ at t_f.

    Make sure to only use the interpolated cubic for t between t_0 and t_f.

    Keyword arguments:
    t_0 -- start time of interpolation
    t_f -- end time of interpolation
    state_0 -- start state [θ₁, θ₂, ω₁, ω₂]ᵀ
    state_f -- end state [θ₁, θ₂, ω₁, ω₂]ᵀ

    Returns:
    coeffs -- 4x2 matrix containing the interpolation coefficients for joint 1
              in column 1 and joint 2 in column 2
    """

    def pos_row(t):
        return np.array([[1, t, t * t, t * t * t]])

    def vel_row(t):
        return np.array([[0, 1, 2 * t, 3 * t * t]])

    lhs = np.concatenate((pos_row(t_0), vel_row(t_0), pos_row(t_f), vel_row(t_f)))
    rhs = np.concatenate((state_0.reshape((2, 2)), state_f.reshape(2, 2)))

    return np.linalg.inv(lhs) @ rhs



"""
Plotting utilities.
"""

import matplotlib.pyplot as plt
import numpy as np


def generate_time_responses(system, refs):
    """
    Generate time-domain responses of the system and the control inputs.

    Keyword arguments:
    system -- system for which to generate responses
    refs -- list of reference vectors, one for each time

    Returns:
    r_rec -- recording of references
    x_rec -- If system.observer exists, a recording of the state estimates.
             Otherwise, a recording of the true states.
    u_rec -- recording of inputs
    y_rec -- recording of outputs
    """
    r_rec = np.zeros((system.x.shape[0], 0))
    x_rec = np.zeros((system.x.shape[0], 0))
    u_rec = np.zeros((system.u.shape[0], 0))
    y_rec = np.zeros((system.y.shape[0], 0))

    # Run simulation
    for i, r in enumerate(refs):
        if i < len(refs) - 1:
            next_r = refs[i + 1]
        else:
            next_r = r
        system.update(r, next_r)

        # Log states for plotting
        r_rec = np.concatenate((r_rec, r), axis=1)
        if hasattr(system, "observer"):
            x_rec = np.concatenate((x_rec, system.observer.x_hat), axis=1)
        else:
            x_rec = np.concatenate((x_rec, system.x), axis=1)
        u_rec = np.concatenate((u_rec, system.u), axis=1)
        y_rec = np.concatenate((y_rec, system.y), axis=1)

    return r_rec, x_rec, u_rec, y_rec


def plot_time_responses(
    state_labels,
    input_labels,
    t_rec,
    r_rec,
    x_rec,
    u_rec,
    title="Time-domain responses",
):
    """Plots time-domain responses of the system and the control inputs.

    Keyword arguments:
    state_labels -- list of state label strings
    input_labels -- list of input label strings
    t_rec -- list of timesteps corresponding to references
    r_rec -- recording of references from generate_time_responses()
    x_rec -- recording of state estimates from generate_time_responses()
    u_rec -- recording of inputs from generate_time_responses()
    title -- title for time-domain plots (default: "Time-domain responses")
    """
    states = x_rec.shape[0]
    inputs = u_rec.shape[0]

    plt.figure()
    subplot_max = states + inputs
    for i in range(states):
        plt.subplot(subplot_max, 1, i + 1)
        if states + inputs > 3:
            plt.ylabel(
                state_labels[i],
                horizontalalignment="right",
                verticalalignment="center",
                rotation=45,
            )
        else:
            plt.ylabel(state_labels[i])
        if i == 0 and title is not None:
            plt.title(title)
        plt.plot(t_rec, r_rec[i, :], label="Reference")
        plt.plot(t_rec, x_rec[i, :], label="State")
        plt.legend()

    for i in range(inputs):
        plt.subplot(subplot_max, 1, states + i + 1)
        if states + inputs > 3:
            plt.ylabel(
                input_labels[i],
                horizontalalignment="right",
                verticalalignment="center",
                rotation=45,
            )
        else:
            plt.ylabel(input_labels[i])
        plt.plot(t_rec, u_rec[i, :], label="Input")
        plt.legend()
    plt.xlabel("Time (s)")



#!/usr/bin/env python3

"""frccontrol example for a single-jointed arm."""

import sys

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

import frccontrol as fct

if "--noninteractive" in sys.argv:
    mpl.use("svg")


class SingleJointedArm:
    """An frccontrol system representing a single-jointed arm."""

    def __init__(self, dt):
        """Single-jointed arm subsystem.

        Keyword arguments:
        dt -- time between model/controller updates
        """
        self.dt = dt

        # Number of motors
        num_motors = 1.0
        # Mass of arm in kg
        m = 2.2675
        # Length of arm in m
        l = 1.2192
        # Arm moment of inertia in kg-m^2
        J = 1 / 3 * m * l**2
        # Gear ratio
        G = 1.0 / 2.0
        self.plant = fct.models.single_jointed_arm(
            fct.models.MOTOR_CIM, num_motors, J, G
        )

        self.sim = self.plant.to_discrete(self.dt)
        self.x = np.zeros((2, 1))
        self.u = np.zeros((1, 1))
        self.y = np.zeros((1, 1))

        # States: angle (rad), angular velocity (rad/s)
        # Inputs: voltage (V)
        # Outputs: angle (rad)
        self.observer = fct.KalmanFilter(
            self.plant, [0.01745, 0.1745329], [0.01], self.dt
        )
        self.feedforward = fct.LinearPlantInversionFeedforward(
            self.plant.A, self.plant.B, self.dt
        )
        self.feedback = fct.LinearQuadraticRegulator(
            self.plant.A, self.plant.B, [0.01745, 0.08726], [12.0], self.dt
        )

        self.u_min = np.array([[-12.0]])
        self.u_max = np.array([[12.0]])

    def update(self, r, next_r):
        """
        Advance the model by one timestep.

        Keyword arguments:
        r -- the current reference
        next_r -- the next reference
        """
        # Update sim model
        self.x = self.sim.A @ self.x + self.sim.B @ self.u
        self.y = self.sim.C @ self.x + self.sim.D @ self.u

        self.observer.predict(self.u, self.dt)
        self.observer.correct(self.u, self.y)
        self.u = np.clip(
            self.feedforward.calculate(next_r)
            + self.feedback.calculate(self.observer.x_hat, r),
            self.u_min,
            self.u_max,
        )


def main():
    """Entry point."""

    dt = 0.005
    single_jointed_arm = SingleJointedArm(dt)

    ts, xprof, vprof, _ = fct.generate_trapezoid_profile(
        max_v=0.5, time_to_max_v=0.5, dt=dt, goal=1.04
    )

    # Run simulation
    refs = [np.array([[xprof[i]], [vprof[i]]]) for i in range(len(ts))]
    r_rec, x_rec, u_rec, _ = fct.generate_time_responses(single_jointed_arm, refs)

    fct.plot_time_responses(
        ["Angle (rad)", "Angular velocity (rad/s)"],
        ["Voltage (V)"],
        ts,
        r_rec,
        x_rec,
        u_rec,
    )
    if "--noninteractive" in sys.argv:
        plt.savefig("single_jointed_arm_response.svg")
    else:
        plt.show()


if __name__ == "__main__":
    main()
